# Filters added to this controller will be run for all controllers in the application.
# Likewise, all the methods added will be available for all controllers.

require_dependency 'utils'
require 'systemu'
require 'net/http'
require 'net/https'
require 'pp'
require 'i18n'

class ApplicationController < ActionController::Base

  layout Proc.new { |controller| controller.request.xhr? ? nil : 'default' }

  prepend_before_filter :load_license

  before_filter :validate_host

  before_filter :format_request
  before_filter :set_cache_buster
  before_filter :logout_on_invalid_entitlement
  before_filter :redirect_to_license_edit_on_invalid_license
  before_filter :intercept_mobile_agents
  before_filter :set_locale

  after_filter  :flash_to_headers

  filter_parameter_logging :password
  protect_from_forgery
  helper_method :current_user, :scope, :logged_in?, :page, :has_next_page?, :vanity_url

  rescue_from Exception, :with => :handle_exception if Rails.env.production?

  def connect_disabled?
    (current_user && current_user.disable_connect?) || session[:disable_connect]
  end
  helper_method :connect_disabled?

  # The function check if transfer is enabled
  # either through connect or connectless (http-gwy)
  # @return [boolean]
  def transfer_enabled?
    !connect_disabled? || CConfiguration.is_gateway_enabled?
  end
  helper_method :transfer_enabled?

  def default_hostname
    @default_host ||= FASPEX_CONFIG['Hostname']
  end

  def accepted_hosts
    @accepted_hosts ||= EConfiguration.get_server_whitelist + [default_hostname, '127.0.0.1', 'localhost']
  end

  def validate_host
    unless accepted_hosts.include?(request.host)
      render text: 'Invalid hostname', status: 404
    end
  end

  # Sets the language of the page based on browser language settings, url parameters, or user database storage.
  # no @return and no @params
  def set_locale
    locale = extract_locale
    I18n.locale =
        if current_user && current_user[:locale] != "none"
          current_user[:locale].to_sym
        elsif valid_locale?(locale)
          locale
        else
          params[:locale] || I18n.default_locale
        end
  end

  def format_request
    # see the implementation of ActionController::Request#format
    #     at http://api.rubyonrails.org/v2.3.8/
    if !request.parameters[:format] &&
        ActionController::Base.use_accept_header &&
        (request.accepts.first == Mime::Type.lookup('*/*')) &&
        !request.content_type.nil?
      request.parameters[:format] = request.content_type.to_sym.to_s
    end
  end

  def handle_exception(e)
    case e
    when ActiveRecord::RecordNotFound
      logger.info("Record not found: #{e}\n#{e.backtrace.join("\n")}")
      redirect_to page_not_found_url
    when ActionController::RoutingError
      logger.info("Invalid URL: #{e}\n#{e.backtrace.join("\n")}")
      redirect_to page_not_found_url
    else
      logger.error("INTERNAL SERVER ERROR: #{e}\n#{e.backtrace.join("\n")}")
      redirect_to internal_server_error_url
    end
  end

  def scope
    nil
  end

  def set_cache_buster
    response.headers["Cache-Control"] = "no-cache, no-store, max-age=0, must-revalidate"
    response.headers["Pragma"]        = "no-cache"
    response.headers["Expires"]       = "Fri, 01 Jan 1990 00:00:00 GMT"
    response.headers['Access-Control-Allow-Origin'] = root_url
  end

  def flash_to_headers
    return unless request.xhr?
    flash_json = Hash[flash.map{|k,v| [k,ERB::Util.h(v)] }].to_json
    response.headers['X-Flash-Messages'] = flash_json
    flash.discard
  end

  # whether to use HTTP basic auth (API)
  def api_request?
    if !request.format.html? && !request.format.js? && !request.xhr?
      if ios_user_agent?
        true
      else
        request.env['HTTP_ACCEPT'].present? && !request.env['HTTP_ACCEPT'].include?('*/*')
      end
    end
  end

  def ios_user_agent?
    if user_agent = request.env['HTTP_USER_AGENT']
      user_agent.downcase!
      user_agent.index('ipod') || user_agent.index('ipad') || user_agent.index('iphone')
    else
      false
    end
  end

  def current_user
    if api_request?
      @current_user ||= oauth_user || basic_auth_user || external_user
    else
      @current_user ||= AuthorizableUser.active.find_by_id(session[:login_user_id]) || external_user
    end
  end

  def external_user
    if !params[:passcode].nil?
      recipient_delivery = RecipientDelivery.find_by_send_to_email_passcode(params[:passcode])
      if recipient_delivery
        return recipient_delivery.recipient
      end
    end
    nil
  end

  def vanity_url user_or_dropbox
    user_or_dropbox.kind_of?(User) ? user_vanity_url(user_or_dropbox) : dropbox_vanity_url(user_or_dropbox)
  end

  def public_url_home_server
    if HomeServer.multi_server?
      EConfiguration.first.public_url_home_server || HomeServer.master_server
    else
      HomeServer.this_server
    end
  end

  def basic_auth_user
    user = authenticate_with_http_basic do |username, password|
      AuthorizableUser.authenticate(username, password, request.remote_ip)
    end
    if user.nil? || !user.errors.empty? || !user.login_from_ip_allowed?(request.remote_ip)
      nil
    else
      user
    end
  end

  def oauth_user
    if request.headers["Authorization"] =~ /^Bearer (.+)/
      access_token = $1
      token = OauthToken.find_by_access_token(access_token)
      if token && !token.expired?
        user = token.user
        return (user && !user.pending? && user.active? && !user.locked?) ? user : nil
      end
    end
    nil
  end

  # UUID for use in connect transfers - unique per user, lazy initialized
  def connect_app_id
    session[:connect_app_id] ||= SecureRandom.uuid
  end
  helper_method :connect_app_id

  def reset_session(reset_login_uuid=true)
    # Do not reset session for 'invalid authenticity token' issues on API requests
    return if caller.first.include?('handle_unverified_request') && (api_request? || request.xhr?)

    notice = flash[:notice]
    current_user.update_attribute(:login_uuid, nil) if current_user.present? && reset_login_uuid

    # Keep track of whether or not connect is disabled for external users
    connect_disabled = session[:disable_connect]

    # Parent method doesn't take any arguments - must explictly call with no args
    super()

    session[:disable_connect] = connect_disabled if connect_disabled
    @current_user = nil

    # reset_session clears flash messages which can be undesirable for logout events such as timeouts
    flash.now[:notice] = notice
  end

  def logged_in?
    current_user.present?
  end

  # TODO display error if exception occurs
  def mail_welcome_notification(user)
    if user.active?
      # send welcome e-mail
      BackgroundNotifier.deliver_welcome_notification(user)
      user.aspera_e_welcome_email_sent = true
      user.save
    end
  rescue
    Utils.log_error("Error delivering welcome notification e-mail to #{user.id}", true)
    return
  end

  def h(text)
    ERB::Util.h(text)
  end

  def app_helper
    Helper.instance
  end

  class Helper
    include Singleton
    include ApplicationHelper
  end

  def check_local_admin_allowed
    if MultiServer::Synchronizer.slave_server?
      redirect_to access_denied_path, :notice => 'Action is not allowed on this server'
    end
  end

  def check_master
    unless HomeServer.this_server.is_master || params[:btnDisconnect]
      redirect_to access_denied_path  and return false
    end
  end

  def render_response(status)
    xml =
        case status
          #login-related
        when 207 then { :message => 'Successfully logged in.' }.to_xml(:root => 'login-response')
          #208: package successfully created, rendered in e_send_controller
        when 401 then { :message => 'Unauthorized.' }.to_xml(:root => 'login-error')
        when 418 then { :message => 'Invalid username/password.' }.to_xml(:root => 'login-error')
        when 419 then { :message => 'Wrong home server.' }.to_xml(:root => 'login-error')
        when 420 then { :message => 'Account inactivated.' }.to_xml(:root => 'login-error')
        when 425 then { :message => 'User password has expired.' }.to_xml(:root => 'login-error')
        when 426 then { :message => 'Package already completed uploading.' }.to_xml(:root => 'status-error')
        when 508 then { :message => 'Only admins allowed to login at this time.' }.to_xml(:root => 'login-error')
        when 509 then { :message => 'Database version error.' }.to_xml(:root => 'login-error')
        when 510 then { :message => "Can't login from this IP address." }.to_xml(:root => 'login-error')
          #send package-related
        when 421 then { :message => 'Invalid recipient list.' }.to_xml(:root => 'send-error')
        when 422 then { :message => 'Package title missing.' }.to_xml(:root => 'send-error')
        when 423 then { :message => 'No files to upload.' }.to_xml(:root => 'send-error')
        when 450 then { :message => 'Bad cookie.' }.to_xml(:root => 'cookie-error')
        when 510 then { :message => 'Error creating package.' }.to_xml(:root => 'send-error')
        when 511 then { :message => 'Error creating upload security token.' }.to_xml(:root => 'send-error')
        when 512 then { :message => 'Error creating fasp URL.' }.to_xml(:root => 'send-error')
          #package status-related
          #209: package status found, rendered in e_package_controller
        when 424 then { :message => "Package UUID (#{params[:pkg_uuid]}) does not match any package." }.to_xml(:root => 'status-error')
        when 451 then { :message => "Page not found. [#{request.path}]" }.to_xml(:root => 'page-not-found')
        end
    render :xml => xml, :status => status
  end

  def page
    page_num = params[:page].to_i
    page_num < 1 ? 1 : page_num
  end

  def package_list_params(params)
    params[:page] = params[:page].to_i == 0 ? nil : params[:page].to_i

    {
        :a_order => sanitize(params[:a_order]),
        :a_sort => sanitize(params[:a_sort]),
        :page => params[:page],
        :search => sanitize(params[:search])
    }.delete_if{|k,v| v.blank?}
  end

  def has_next_page?(will_paginate_collection)
    will_paginate_collection.total_pages > page
  end

  def transfer_host
    Share.default_destination.node.get_server_address request.remote_ip, request.host
  end

  def redirect_to(options = {}, response_status = {})
    if request.xhr?
      render(:update) { |page| page.redirect_to(options) }
    else
      super(options, response_status)
    end
  end

  def linux?
    return if Aspera::OS.linux?
    redirect_to access_denied_path
  end

  def automatically_login_using_saml_idp?
    # Automatically log in using SAML if a default has been specified.
    @saml_enabled ||= SamlConfiguration.default_saml_set? || SamlConfiguration.select_by_domain_url(root_url)
  end

  def saml_login_by_domain?
    SamlConfiguration.select_by_domain_url(root_url)
  end

  def after_logout_url
    if current_user and current_user.saml? && SamlConfiguration.find(current_user.saml_configuration_id).redirect_on_logout?
      saml_logout_url(id: current_user.saml_configuration_id)
    elsif automatically_login_using_saml_idp?
      new_login_url(local: true)
    else
      new_login_url
    end
  end

  def filter_params_for_deleted_user_profile_fields
    user_profile_data_attributes_params = params[:user][:user_profile_data_attributes]
    user_profile_data_attributes_params.each do |k, v|
      user_profile_field_id = v["user_profile_field_id"].to_i
      unless (UserProfileField.all.map(&:id).include?(user_profile_field_id) && UserProfileField.find_by_id(user_profile_field_id).enabled?)
        user_profile_data_attributes_params.delete(k)
      end
    end
    user_profile_data_attributes_params
  end

  private

  def request_url
    request.protocol + request.host + (request.port ? ":#{request.port}" : '') + EConfiguration.get_url_namespace
  end

  def public_url_request?
    request_url == public_url_home_server.url
  end

  def user_vanity_url user
    if public_url_request?
      url = public_url_home_server.url + public_user_url_path(user.friendly_id, skip_relative_url_root: true)
    else
      url = request_url + public_user_url_path(user.friendly_id, skip_relative_url_root: true)
    end
    URI.unescape url
  end

  def dropbox_vanity_url dropbox
    if public_url_request?
      url = public_url_home_server.url + public_dropbox_url_path(dropbox.friendly_id, skip_relative_url_root: true)
    else
      url = request_url + public_dropbox_url_path(dropbox.friendly_id, skip_relative_url_root: true)
    end
    URI.unescape url
  end

  def save_intended_destination
    #save info on where he intended to go (used after it is found that the user has not logged in yet)
    session[:intended_params] = params if !request.post? && !request.xhr?
  end

  def verify_user_is_admin
    return if current_user.present? && current_user.admin?
    redirect_to access_denied_path
  end

  def current_user_is_server_admin
    current_user.present? && current_user.admin_with_privilege?(User::PERM_SERVER)
  end
  def verify_user_is_server_admin
    return if current_user_is_server_admin
    redirect_to access_denied_path
  end

  def redirect_to_license_edit_on_invalid_license
    if !EConfiguration.safe_net_entitlement_mode_on? && !valid_license?
      if api_request?
        raise Api::Error.new(403, 'Invalid license', 'Invalid license')
      else
        redirect_to edit_admin_settings_license_url and return
      end
    end
  end

  def logout_on_invalid_entitlement
    if EConfiguration.safe_net_entitlement_mode_on?
      entitlement = Entitlement.new
      if entitlement.stale_data? || !EConfiguration.entitled?
        if current_user
          if !current_user.super_admin?
            logger.warn("Entitlement Failure:#{entitlement.stale_data? ? 'Stale Entitlement Date' : 'App Not Entitled'} - Logging Out user [#{current_user.id}-#{current_user.name}]")
            reset_session
            save_intended_destination
            redirect_to after_logout_url
          else
            entitlement_error_message = entitlement.stale_data? ? 'Stale Entitlement Data' : 'Invalid Entitlement License'
            logger.warn("Entitlement Failure:#{entitlement_error_message} - Redirecting user [#{current_user.id}-#{current_user.name}]")
            flash[:error] = entitlement_error_message
            redirect_to edit_admin_settings_entitlement_url
          end
        end
      end
    end
  end

  def valid_license_or_entitlement?
    if !EConfiguration.safe_net_entitlement_mode_on?
      valid_license?
    else
      if Entitlement.new.stale_data? || !EConfiguration.entitled?
        return false
      else
        return true
      end
    end
  end

  def valid_license?
    begin
      AsperaLicenseReader.load_license.valid?
    rescue Exception
      Utils.log_info("Error reading from license file: #{$!.to_s}")
      Utils.log_debug("#{$!.backtrace.join("\n")}")
      return false
    end
  end

  def load_license
    AsperaLicenseReader.load_license unless EConfiguration.safe_net_entitlement_mode_on?
  rescue Exception
    Utils.log_info("Error reading from license file: #{$!.to_s}")
    Utils.log_debug("#{$!.backtrace.join("\n")}")
  end

  def sanitize(str)
    ActionController::Base.helpers.sanitize(str)
  end

  def sanitize_sort_parameters
    params[:a_order] = (params[:a_order].downcase == 'a' || params[:a_order].downcase == 'd' ? params[:a_order].downcase : 'a') if params[:a_order].present?
    params[:page] = params[:page].to_i if params[:page].present?

    # Sanitize and scrub any quotes
    [:a_sort, :a_order, :search].map{ |p| params[p] = sanitize(params[p]).to_s.gsub(/["']/, '').presence if params[p].present? }

    # Remove HTML tags and URLs from the search term received from the user
    params[:search] = Utils.sanitize_and_strip_tags(params[:search])
  end

  def sanitize_html(val)
    sanitize(val.to_s.gsub(/["'<>]/, ''))
  end

  # Recursively sanitize parameter hash
  def sanitize_hash(val, remove_html=false)
    if val.class == Hash || val.class == HashWithIndifferentAccess
      sanitized_hash = {}

      val.map{ |k, v| sanitized_hash[k] = sanitize_hash(v, remove_html) }

      sanitized_hash
    elsif val.class == Array
      val.map{ |v| sanitize_hash(v, remove_html) }
    elsif val.class == String
      remove_html ? sanitize_html(val) : sanitize(val)
    else
      val
    end
  end

  def intercept_mobile_agents
    if (request.user_agent =~ /Android|iPod|iPad|iPhone/) && params[:skipmobile].nil? && ((params[:controller] == 'received') || (params[:controller] == 'external_delivery'))
      url = (request.fullpath.starts_with?("https") ? request.fullpath : request.url)
      user_delivery = UserDelivery.find_by_id(params[:id])
      delivery_id = user_delivery.present? ? user_delivery.delivery_id.to_s : ''

      @app_url = mobile_package_received_url(params[:id]) + '?' + (request.query_string.present? ? request.query_string + '&' : '') + 'delivery=' + delivery_id

      @app_url = @app_url.sub(/https?/, 'faspex-pkg')

      @skip_to_web_url = url
      @app_store_url = request.user_agent =~ /Android/ ? "https://play.google.com/store/apps/details?id=com.asperasoft.faspex" : "https://itunes.apple.com/app/id516385262"
      render 'mobile/landing', :layout => 'mobile_landing'
      return false
    end
  end

  # Checks to see if extracted or parsed locales match with the accepted locales of our application.
  # @param locale [String] two letter encoding for I18n
  # @return [boolean] for whether or not this locale is supported
  def valid_locale?(locale)
    I18n.available_locales.map(&:to_s).include?(locale)
  end

  # Extracts the locale of the browser from the HTTP request environmental variables and produces an I18n readable encoding.
  # @param none
  # @return the first accepted language listed in the browser encoded into two letters
  def extract_locale
    acc_lang = request.env["HTTP_ACCEPT_LANGUAGE"]
    return unless acc_lang
    acc_lang.scan(/^[a-z]{2}/).first
  end
end
