# including controller must define downloadable_object, which returns a delivery or recipient_delivery
# to be downloaded
module DownloadableActions

  def self.included base
    base.class_eval do
      before_filter :find_delivery_and_package, only: [:download, :forward, :forward_submit]
      before_filter :ensure_package_downloadable!, only: :download
      include InstanceMethods
    end
  end

  module InstanceMethods

    def download
      create_transfer_url
      # If delivery.prevent_http_download, then force use of connect.
      if @delivery.prevent_http_download
         params[:transfer_type] = "connect"
      end
      case params[:transfer_type]
        when 'http'
          render 'deliveries/http_download'
        when 'http_gateway'
          render 'deliveries/http_gateway_download'
        when 'connect'
          render 'deliveries/connect_download'
        end
    rescue => e
      flash[:error] = "Download failed, please contact your administrator."
      logger.error("Unable to get transfer spec: #{e}\n#{e.backtrace.join("\n")}")
      render 'deliveries/error'
    end

    def find_delivery_and_package
      @delivery = downloadable_object.delivery
      @package  = downloadable_object.e_package
    rescue => e
      respond_to do |format|
        format.html do
          raise
        end
        format.json do
          render_json_error(:not_found, e.message)
        end
      end
    end

    private

    def create_transfer_url
      # TODO: Handle the case where the storage node has been deleted
      @spec = downloadable_object.get_transfer_spec ESession::DIRECTION_DOWNLOAD,
        params[:items],
        current_user,
        token: true,
        host: transfer_host
      download_spec_hash = JSON.parse(@spec.to_json)

      #This condition only applies to HTTP gateway as the
      # gateway needs 'paths' to create folders for download
      # 'source_root' and 'paths' are concatentated to locate the file
      # While connect handles all such cases
      zip_required = false
      if params[:transfer_type] == 'http_gateway'
        file_paths = []

        if params[:relative_path] == "" || params[:relative_path] == "/"
          source_root = @package.inner_directory_path
        else
          source_root = File.join(@package.inner_directory_path, params[:relative_path])
        end

        get_nested_paths download_spec_hash["paths"], file_paths, source_root
        zip_required = is_zip_required download_spec_hash["paths"]
        download_spec_hash["paths"] = file_paths
        download_spec_hash["source_root"] = source_root
      end

      download_spec_hash["authentication"] = "token"
      download_spec_hash["tags"]["aspera"]["faspex"]["requestor"].reject! {|k, v| k=="name"}  if current_user.is_a?(ExternalUser) && FASPEX_CONFIG['HideSenderUsernameToExternalRecipients']
      download_spec_hash.reject! {|k, v| v.blank?}
      @download_spec_hash = download_spec_hash
      @http_gateway_url = CConfiguration.first.http_gateway_url if params[:transfer_type] == 'http_gateway'
      @download_name = generate_download_name params[:items], @package
      download_spec_hash["download_name"] = @download_name
      download_spec_hash["zip_required"] = zip_required
      download_spec_hash["multi_session"] = 0
    end


    # This function calls NODE API /files/browse recursively for each path until
    # the type of the path is NOT directory.
    def get_nested_paths(source_paths, file_paths, source_root)
      source_paths.each do |source_path|
        path = source_path["source"]
        get_path path, file_paths, source_root
      end
    end

    # The method get_path is recursively called with the parameter 'path' to resolve and
    # after resolving the 'path' to a fully qualified file path its stored in the array 'file_paths'
    # This function calls NODE API /files/browse recursively for each path until
    # the type of the path is NOT directory.
    def get_path(path, file_paths, source_root)
      # setting the count to 10000 to override the default count value of 100 in the file browse call. This will allow us to get the path of
      # all the files in the directory if the count is greater than 100. This limitation will be documented in the admin docs. FASPX-1586
      node_api_response = @delivery.storage_share.node.api_client.browse(path, {:count => 10000})
      if node_api_response.error?
        @directory_items = []
      else
        @directory_items = node_api_response.directory_items
        # This scenario will happen when the path is a file
        # and NOT a directory, the node API returns empty items array but one item in self.
        # Ideally this should be fixed in NodeApi::Response::Browse where directory items
        # are created.
        if @directory_items.length == 0
          # The substring is done for http gateway to create folder structure
          # The values passed in "paths" is used for creating folder structure
          # and the value "source_root" is the parent directory
          file_path = path.sub(source_root, "")
          file_paths.push({"source" => file_path})
        end
      end

      @directory_items.each do |directory_item|
        if directory_item.type != "directory"
          file_path = directory_item.path.sub(source_root, "")
          file_paths.push({"source" => file_path})
        else
          get_path directory_item.path, file_paths, source_root
        end
      end
    end

    def transfer_host
      downloadable_object.default_transfer_host(request) rescue downloadable_object.delivery.default_transfer_host(request)
    end

    def ensure_package_downloadable!
      redirect_to access_denied_url unless @package.can_be_downloaded_by?(current_user, request.remote_ip)
    end

    def generate_download_name items, package
      # If the item size is 1 - it can either be file, folder, or entire package
      # So we check for if the path is root if not take the base name of the path
      if items.size == 1 && items[0] != "/"
        path = Pathname.new(items.first)
        name = path.basename.to_s
      else
        name = package.inner_directory_name
      end
      name
    end

    # This function check if zip is required or not
    # It takes the arg `return true`
    # @argument source_paths to be downloaded
    # @return [boolean]
    def is_zip_required source_paths
      if source_paths.size == 1
        node_api_response = @delivery.storage_share.node.api_client.browse(source_paths[0]["source"])
        if node_api_response.error?
          logger.error "Error in checking the path for file/folder. Defaulting the condition to zip required (return true) "
          return true
        else
          @directory_items = node_api_response.directory_items
          if @directory_items.length == 0
            return false
          else
            return true
          end
        end
      else
        return true
      end
    end

    def partial? items, package
      !package.entire? items
    end
  end
end
